/** \file rpi_cava.h
 *  \brief This module contains continuous tasks for cava (spectrum analyzer) comunication
 */

/***************************************************************************************************
 * Includes
 **************************************************************************************************/
#include "rpi_cava.h"

#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <pthread.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/***************************************************************************************************
 * Preprocessor Constants and Macros
 **************************************************************************************************/

/**
 * Constant for Maximum char in a line (128 number of (5 char + 1 delim) + \n )
 */
#define MAXLINECHAR 128 * (5 + 1) + 1

/***************************************************************************************************
 * Type and Contant Definitions
 **************************************************************************************************/

/***************************************************************************************************
 * Persistent Variables
 **************************************************************************************************/

/**
 * Cava process id
 */
pid_t cavaPid;
/**
 * Boolean for interrupting the main thread loop
 */
bool isFifoReaderRunning = false;
/**
 * Task thread identifier
 */
pthread_t fifoReader;

int fdCavaInput;
char lineBuffer[MAXLINECHAR];
uint16_t buffer[128 + 2];

/***************************************************************************************************
 * Internal Function Prototypes
 **************************************************************************************************/

/**
 * \brief This function is used to thread main execution function
 *
 * \param   arg not used.
 *
 * \return  NULL.
 */
static void *fifo_to_buffer(void *arg);

/***************************************************************************************************
 * External Function Definitions
 **************************************************************************************************/

void setup_cava() {
  if ((cavaPid = fork()) == -1) {
    perror("fork");
    exit(1);
  }

  if (cavaPid == 0) {
    /* Child process*/
    char *args[] = {"/usr/local/bin/cava", "-p", "/home/pi/LedBars/RpiLedBars/cava_config", NULL};

    if (execv(args[0], args) != 0) {
      perror("execv");
    }

  } else {
    sleep(1);
  }
}

void close_cava() {
  stop_cava_bg_worker();
  kill(cavaPid, SIGTERM);
}

void start_cava_bg_worker() {
  isFifoReaderRunning = true;
  pthread_create(&fifoReader, NULL, fifo_to_buffer, NULL);
}

void stop_cava_bg_worker() {
  isFifoReaderRunning = false;
  if (pthread_join(fifoReader, NULL) != 0) {
    perror("pthread_join");
  }
}

int get_cava_buffer(uint16_t **buffer_dst) {
  *buffer_dst = buffer;
  return 0;
}

/***************************************************************************************************
 * Internal Function Definitions
 **************************************************************************************************/

static void *fifo_to_buffer(void *arg) {
  struct pollfd fds[1];
  int timeoutMs;
  int ret;
  size_t valueIndex = 0, charOffset = 0;
  char strValue[6] = "0\0";
  bool hasToBeDiscarded = true;

  if start_cava_process ()
    ;
  if ((fdCavaInput = open("/tmp/cava_output", O_RDONLY | O_NONBLOCK)) < 0) {
    perror("open");
    close_cava();
    exit(1);
  }

  memset(fds, 0, sizeof(fds));
  fds[0].fd = fdCavaInput;
  fds[0].events = POLLIN;
  timeoutMs = 10;

  while (isFifoReaderRunning) {

    if ((ret = poll(fds, 1, timeoutMs)) == 1) {
      int nread;
      nread = read(fdCavaInput, lineBuffer, 128 + 1);

      if (nread >= 0) {
        for (size_t i = 0; i < nread; ++i) {
          char current = lineBuffer[i];

          if (hasToBeDiscarded) {
            if (current == '\n') {
              charOffset = 0;
              strValue[charOffset] = '\0';
              valueIndex = 0;
              hasToBeDiscarded = false;
            }

          } else {
            if ('0' <= current && current <= '9') {
              strValue[charOffset++] = current;
            } else if (current == '\n' || current == ';') {
              strValue[charOffset] = '\0';
              charOffset = 0;
              buffer[valueIndex++] = atoi(strValue);

              if (current == '\n' || valueIndex > 129) {
                valueIndex = 0;

                if (valueIndex > 129) {
                  fprintf(stderr, "Buffer overflow, \\n missed, discarding next\n");
                  hasToBeDiscarded = true;
                }
              }

            } else {
              fprintf(stderr, "Unexpected char %d [%c]\n", current, current);
            }
          }
        }
      } else {
        if (errno != EAGAIN) {
          perror("read");
        }
      }
    } else if (ret < 0) {
      fprintf(stderr, "error polling %d: %s\n", fdCavaInput, strerror(errno));
    }
  }

  close(fdCavaInput);

  return NULL;
}

static int start_cava_process() {
  int fdCavaPipe[2];

  if (pipe(fdCavaPipe) < 0) {
    fprintf(stderr, "Cava pipe failure: %s\n", strerror(errno));
    return -EXIT_FAILURE;
  }

  if ((cavaPid = fork()) < 0) {
    fprintf(stderr, "Cava fork failure: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }

  if (cavaPid == 0) {
    /* Child process*/
    char *args[] = {"cava", "-p", "/home/pi/LedBars/RpiLedBars/cava_config", NULL};

    /* Close reading end of the pipe */
    close(fdCavaPipe[0]);
    /* Dup writing end of the pipe in place of stdout */
    dup2(fdCavaPipe[1], STDOUT_FILENO);
    /* Close writing end of the pipe */
    close(fdCavaPipe[1]);

    execvp(args[0], args);
    fprintf(stderr, "Cava execvp failure or return: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  } else {
    /* Close writing end of the pipe */
    close(fdCavaPipe[1]);
    /* Set reading end of the pipe non-blocking */
    flags = fcntl(fdCavaPipe[0], F_GETFL, 0);
    fcntl(fdCavaPipe[0], F_SETFL, flags | O_NONBLOCK);
    /* Return reading end of the pipe */
    return fdCavaPipe[0];
  }

  /* Unreachable  */
  return -1;
}

static void stop_cava_process() { kill(cavaPid, SIGTERM); }
